/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.model.framestore;

import java.lang.reflect.*;
import java.util.*;

import edu.stanford.smi.protege.model.*;

public class ArgumentCheckingFrameStoreHandler extends AbstractFrameStoreInvocationHandler {

    public Object handleInvoke(Method m, Object[] args) {
        checkArguments(m, args);
        return invoke(m, args);
    }

    private void checkArguments(Method m, Object[] args) {
        if (args != null) {
            for (int i = 0; i < args.length; ++i) {
                Object arg = args[i];
                checkArgument(m, i, arg);
            }
        }
    }

    private void checkArgument(Method m, int position, Object arg) {
        if (arg == null && !allowedNull(m, position)) {
            throw new IllegalArgumentException("Null argument for method " + m + " at position: " + position);
        } else if (arg instanceof Frame) {
            checkFrame(m, position, (Frame) arg);
        } else if (arg instanceof Collection) {
            checkCollection(m, position, (Collection) arg);
        }
    }

    private void checkFrame(Method m, int position, Frame arg) {
    }

    private void checkValue(Method m, int position, Object value) {
        if (value == null) {
            throw new IllegalArgumentException("Null collection element for method " + m + " at position: " + position);
        } else if (!isValidClass(value)) {
            throw new IllegalArgumentException(
                "Invalid collection element " + value + " for method " + m + " at position: " + position);
        }
    }

    private boolean isValidClass(Object value) {
        return value instanceof Boolean
            || value instanceof Float
            || value instanceof Integer
            || value instanceof Frame
            || value instanceof String;
    }

    private void checkCollection(Method m, int position, Collection arg) {
        Iterator i = arg.iterator();
        while (i.hasNext()) {
            Object o = i.next();
            checkValue(m, position, o);
        }
    }

    private boolean allowedNull(Method m, int position) {
        String name = m.getName();
        return (name.startsWith("create") && (position == 0 || position == 1))
            || (name.startsWith("setAssociated") && position == 1)
            || (name.startsWith("setInverseSlot") && position == 1);
    }
}

/*
public class ArgumentCheckingFrameStore extends AbstractDelegatingFrameStore {
    public static class NullArgumentException extends IllegalArgumentException {}
    public static class DeletedFrameException extends IllegalArgumentException {}
    public static class BadTypeException extends IllegalArgumentException {}

    private void checkFrame(Frame frame) {
        if (frame == null) {
            throw new NullArgumentException();
        }
        if (frame.getKnowledgeBase() == null) {
            throw new DeletedFrameException();
        }
    }
    private void checkCls(Cls cls) {
        checkFrame(cls);
    }
    private void checkSlot(Slot slot) {
        checkFrame(slot);
    }
    private void checkSlotOrNull(Slot slot) {
        if (slot != null) {
            checkSlot(slot);
        }
    }
    private void checkFacet(Facet facet) {
        checkFrame(facet);
    }
    private void checkFacetOrNull(Facet facet) {
        if (facet != null) {
            checkFacet(facet);
        }
    }
    private void checkSimpleInstance(SimpleInstance simpleInstance) {
        checkFrame(simpleInstance);
    }
    private void checkInstance(Instance instance) {
        checkFrame(instance);
    }
    private void checkClses(Collection clses) {
        checkCollection(clses, Cls.class);
    }
    private void checkSlots(Collection slots) {
        checkCollection(slots, Slot.class);
    }
    private void checkCollection(Collection collection, Class clas) {
        Iterator i = collection.iterator();
        while (i.hasNext()) {
            Object o = i.next();
            if (!clas.isInstance(o)) {
                throw new BadTypeException();
            }
        }
    }
    private void checkStringOrNull(String s) {
        // do nothing
    }
    private void checkString(String s) {
        if (s == null) {
            throw new NullArgumentException();
        }
    }
    private void checkValue(Object value) {
        if (value == null) {
            throw new NullArgumentException();
        }
    }
    private void checkValues(Collection values) {
        checkValue(values);
        if (values.contains(null)) {
            throw new NullArgumentException();
        }
    }
    private void checkSlotValue(Slot slot, Object value) {
        checkValue(value);
        // @todo check type
    }
    private void checkFacetValue(Facet facet, Object value) {
        checkValue(value);
        // @todo check type
    }
    private void checkOwnSlotValue(Frame frame, Slot slot, Object value) {
        checkValue(value);
        // @todo check type
    }
    private void checkOwnSlotValues(Frame frame, Slot slot, Collection values) {
        checkValues(values);
        // @todo check type & cardinality
    }
    private void checkTemplateSlotValues(Frame frame, Slot slot, Collection values) {
        checkValue(values);
        // @todo check type and cardinality
    }
    private void checkTemplateFacetValues(Frame frame, Slot slot, Facet facet, Collection values) {
        checkValue(values);
        // @todo check type and cardinality
    }
    private void checkQuery(Query query) {
        checkValue(query);
    }

    public Cls createCls(String name, Collection directTypes, Collection directSuperclasses, Session session) throws Exception {
        checkStringOrNull(name);
        checkClses(directTypes);
        checkClses(directSuperclasses);
        return getDelegate().createCls(name, directTypes, directSuperclasses, session);
    }

    public void deleteFrame(Frame frame, Session session) throws Exception {
        checkFrame(frame);
        getDelegate().deleteFrame(frame, session);
    }
    public SimpleInstance createSimpleInstance(String name, Collection directTypes, Session session) throws Exception {
        checkStringOrNull(name);
        checkClses(directTypes);
        return getDelegate().createSimpleInstance(name, directTypes, session);
    }
    public Slot createSlot(String name, Collection directTypes, Collection directSuperslots, Session session) throws Exception {
        checkStringOrNull(name);
        checkClses(directTypes);
        checkSlots(directSuperslots);
        return getDelegate().createSlot(name, directTypes, directSuperslots, session);
    }
    public Facet createFacet(String name, Collection directTypes, Session session) throws Exception {
        checkStringOrNull(name);
        checkClses(directTypes);
        return getDelegate().createFacet(name, directTypes, session);
    }
    public void setDirectOwnSlotValues(Frame frame, Slot slot, Collection values, Session session) throws Exception {
        checkFrame(frame);
        checkSlot(slot);
        checkOwnSlotValues(frame, slot, values);
        getDelegate().setDirectOwnSlotValues(frame, slot, values, session);
    }
    public void setDirectTemplateSlotValues(Cls cls, Slot slot, Collection values, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        checkTemplateSlotValues(cls, slot, values);
        getDelegate().setDirectTemplateSlotValues(cls, slot, values, session);
    }
    public void setDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet, Collection values, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        checkFacet(facet);
        checkTemplateFacetValues(cls, slot, facet, values);
        getDelegate().setDirectTemplateFacetValues(cls, slot, facet, values, session);
    }
    public void addDirectTemplateSlot(Cls cls, Slot slot, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        getDelegate().addDirectTemplateSlot(cls, slot, session);
    }
    public void removeDirectTemplateSlot(Cls cls, Slot slot, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        getDelegate().removeDirectTemplateSlot(cls, slot, session);
    }
    public void moveDirectTemplateSlot(Cls cls, Slot slot, int index, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        getDelegate().moveDirectTemplateSlot(cls, slot, index, session);
    }
    public void addDirectSuperclass(Cls cls, Cls superclass, Session session) throws Exception {
        checkCls(cls);
        checkCls(superclass);
        getDelegate().addDirectSuperclass(cls, superclass, session);
    }
    public void removeDirectSuperclass(Cls cls, Cls superclass, Session session) throws Exception {
        checkCls(cls);
        checkCls(superclass);
        getDelegate().removeDirectSuperclass(cls, superclass, session);
    }
    public void moveDirectSubclass(Cls cls, Cls subclass, int index, Session session) throws Exception {
        checkCls(cls);
        checkCls(subclass);
        getDelegate().moveDirectSubclass(cls, subclass, index, session);
    }
    public void addDirectSuperslot(Slot slot, Slot superslot, Session session) throws Exception {
        checkSlot(slot);
        checkSlot(superslot);
        getDelegate().addDirectSuperslot(slot, superslot, session);
    }
    public void removeDirectSuperslot(Slot slot, Slot superslot, Session session) throws Exception {
        checkSlot(slot);
        checkSlot(superslot);
        getDelegate().removeDirectSuperslot(slot, superslot, session);
    }
    public void addDirectType(Instance instance, Cls type, Session session) throws Exception {
        checkInstance(instance);
        checkCls(type);
        getDelegate().addDirectType(instance, type, session);
    }
    public void removeDirectType(Instance instance, Cls type, Session session) throws Exception {
        checkInstance(instance);
        checkCls(type);
        getDelegate().removeDirectType(instance, type, session);
    }

    public Set getClses(Session session) throws Exception {
        return getDelegate().getClses(session);
    }
    public Set getSlots(Session session) throws Exception {
        return getDelegate().getSlots(session);
    }
    public Set getFacets(Session session) throws Exception {
        return getDelegate().getFacets(session);
    }
    public Frame getFrame(String name, Session session) throws Exception {
        checkString(name);
        return getDelegate().getFrame(name, session);
    }
    public Set getFrames(Session session) throws Exception {
        return getDelegate().getFrames(session);
    }
    public Set getOwnSlots(Frame frame, Session session) throws Exception {
        checkFrame(frame);
        return getDelegate().getOwnSlots(frame, session);
    }
    public List getDirectOwnSlotValues(Frame frame, Slot slot, Session session) throws Exception {
        checkFrame(frame);
        checkSlot(slot);
        return getDelegate().getDirectOwnSlotValues(frame, slot, session);
    }
    public Collection getOwnSlotValues(Frame frame, Slot slot, Session session) throws Exception {
        checkFrame(frame);
        checkSlot(slot);
        return getDelegate().getOwnSlotValues(frame, slot, session);
    }
    public Set getTemplateSlots(Cls cls, Session session) throws Exception {
        checkCls(cls);
        return getDelegate().getTemplateSlots(cls, session);
    }
    public List getDirectTemplateSlots(Cls cls, Session session) throws Exception {
        checkCls(cls);
        return getDelegate().getDirectTemplateSlots(cls, session);
    }
    public Set getTemplateSlotValues(Cls cls, Slot slot, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        return getDelegate().getTemplateSlotValues(cls, slot, session);
    }
    public List getDirectTemplateSlotValues(Cls cls, Slot slot, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        return getDelegate().getDirectTemplateSlotValues(cls, slot, session);
    }
    public Set getTemplateFacets(Cls cls, Slot slot, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        return getDelegate().getTemplateFacets(cls, slot, session);
    }
    public Set getTemplateFacetValues(Cls cls, Slot slot, Facet facet, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        checkFacet(facet);
        return getDelegate().getTemplateFacetValues(cls, slot, facet, session);
    }
    public List getDirectTemplateFacetValues(Cls cls, Slot slot, Facet facet, Session session) throws Exception {
        checkCls(cls);
        checkSlot(slot);
        checkFacet(facet);
        return getDelegate().getDirectTemplateFacetValues(cls, slot, facet, session);
    }
    public List getDirectSuperclasses(Cls cls, Session session) throws Exception {
        checkCls(cls);
        return getDelegate().getDirectSuperclasses(cls, session);
    }
    public Set getSuperclasses(Cls cls, Session session) throws Exception {
        checkCls(cls);
        return getDelegate().getSuperclasses(cls, session);
    }
    public List getDirectSubclasses(Cls cls, Session session) throws Exception {
        checkCls(cls);
        return getDelegate().getDirectSubclasses(cls, session);
    }
    public Set getSubclasses(Cls cls, Session session) throws Exception {
        checkCls(cls);
        return getDelegate().getSubclasses(cls, session);
    }
    public List getDirectSuperslots(Slot slot, Session session) throws Exception {
        checkSlot(slot);
        return getDelegate().getDirectSuperslots(slot, session);
    }
    public Set getSuperslots(Slot slot, Session session) throws Exception {
        checkSlot(slot);
        return getDelegate().getSuperslots(slot, session);
    }
    public List getDirectSubslots(Slot slot, Session session) throws Exception {
        checkSlot(slot);
        return getDelegate().getDirectSubslots(slot, session);
    }
    public Set getSubslots(Slot slot, Session session) throws Exception {
        checkSlot(slot);
        return getDelegate().getSubslots(slot, session);
    }
    public List getDirectTypes(Instance instance, Session session) throws Exception {
        checkInstance(instance);
        return getDelegate().getDirectTypes(instance, session);
    }
    public Set getTypes(Instance instance, Session session) throws Exception {
        checkInstance(instance);
        return getDelegate().getTypes(instance, session);
    }
    public List getDirectInstances(Cls cls, Session session) throws Exception {
        checkCls(cls);
        return getDelegate().getDirectInstances(cls, session);
    }
    public Set getInstances(Cls cls, Session session) throws Exception {
        checkCls(cls);
        return getDelegate().getInstances(cls, session);
    }
    public List getEvents(Session session) throws Exception {
        return getDelegate().getEvents(session);
    }
    public Set getReferences(Object value, Session session) throws Exception {
        checkValue(value);
        return getDelegate().getReferences(value, session);
    }
    public Set getMatchingReferences(String value, Session session) throws Exception {
        checkValue(value);
        return getDelegate().getMatchingReferences(value, session);
    }
    public Set getFramesWithOwnSlotValue(Slot slot, Object value, Session session) throws Exception {
        checkSlot(slot);
        checkSlotValue(slot, value);
        return getDelegate().getFramesWithOwnSlotValue(slot, value, session);
    }
    public Set getFramesWithMatchingOwnSlotValue(Slot slot, String value, Session session) throws Exception {
        checkSlot(slot);
        checkString(value);
        return getDelegate().getFramesWithMatchingOwnSlotValue(slot, value, session);
    }
    public Set executeQuery(Query query, Session session) throws Exception {
        checkQuery(query);
        return getDelegate().executeQuery(query, session);
    }

    public String getFrameName(Frame frame, Session session) throws Exception {
        checkFrame(frame);
        return getDelegate().getFrameName(frame, session);
    }

    public void setFrameName(Frame frame, String name, Session session) throws Exception {
        checkFrame(frame);
        checkString(name);
        getDelegate().setFrameName(frame, name, session);
    }

    public Facet getAssociatedFacet(Slot slot, Session session) throws Exception {
        checkSlot(slot);
        return getDelegate().getAssociatedFacet(slot, session);
    }

    public Slot getAssociatedSlot(Facet facet, Session session) throws Exception {
        checkFacet(facet);
        return getDelegate().getAssociatedSlot(facet, session);
    }

    public void setAssociatedFacet(Slot slot, Facet facet, Session session) throws Exception {
        checkSlot(slot);
        checkFacetOrNull(facet);
        getDelegate().setAssociatedFacet(slot, facet, session);
    }

    public Slot getInverseSlot(Slot slot, Session session) throws Exception {
        checkSlot(slot);
        return getDelegate().getInverseSlot(slot, session);
    }

    public void setInverseSlot(Slot slota, Slot slotb, Session session) throws Exception {
        checkSlot(slota);
        checkSlotOrNull(slotb);
        getDelegate().setInverseSlot(slota, slotb, session);
    }
}
*/
