/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.server;

import java.io.*;
import java.net.*;
import java.net.UnknownHostException;
import java.rmi.*;
import java.rmi.registry.*;
import java.rmi.server.*;
import java.util.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.model.framestore.*;
import edu.stanford.smi.protege.resource.*;
import edu.stanford.smi.protege.util.*;

public class Server extends UnicastRemoteObject implements RemoteServer {
    private Map _nameToOpenProjectMap = new HashMap();
    private Map _projectToServerProjectMap = new HashMap();
    private KnowledgeBase _systemKb;
    private Slot _nameSlot;
    private Slot _passwordSlot;
    private Slot _locationSlot;
    private Cls _userCls;
    private Cls _projectCls;
    private List _sessions = new ArrayList();
    private URI _baseURI;
    private Map _sessionToProjectsMap = new HashMap();
    private Thread _updateThread;
    private int registryPort = Registry.REGISTRY_PORT;
    private URI metaprojectURI;
    private static final int NO_SAVE = -1;
    private int _saveIntervalMsec = NO_SAVE;
    private static final String SAVE_INTERVAL_OPTION = "-saveIntervalSec=";
    private static final String NOPRELOAD_OPTION = "-nopreload";
    private static final String OPTION_CHAR = "-";
    private boolean preload = true;

    public static void main(String[] args) {
        try {
            startServer(args);
        } catch (Exception e) {
            System.out.println("Exception occurred: " + e);
        }
    }

    /**
     * Calls startServer with {@link FixedPortRmiSocketFactory} as the socket
     * factory.
     */
    public static void startServer(String[] args) throws IOException {
        startServer(args, new FixedPortRmiSocketFactory());
    }

    /**
     * Start up the server.
     *
     * @param args the arguments to the server
     * @param sf the socket factory to be used
     * @throws IOException if the socket factory has already been set
     * @see RMISocketFactory#setSocketFactory(RMISocketFactory)
     */
    public static void startServer(String[] args, RMISocketFactory sf) throws IOException {
        SystemUtilities.initialize();
        new Server(args, sf);
        System.out.println("Protege server ready to accept connections...");
        System.out.flush();
    }

    public static String getBoundName() {
        return Text.getProgramName();
    }

    protected static String getLocalBoundName() {
        return getBoundName();
    }

    private Registry getRegistry() throws RemoteException, InterruptedException {
        return LocateRegistry.getRegistry(registryPort);
    }

    private void parseArgs(String[] args) {
        for (int i = 0; i < args.length; ++i) {
            parseArg(args[i]);
        }
    }

    protected void parseArg(String arg) {
        if (arg.startsWith(SAVE_INTERVAL_OPTION)) {
            extractSaveInterval(arg);
        } else if (arg.startsWith(NOPRELOAD_OPTION)) {
            preload = false;
        } else if (arg.startsWith(OPTION_CHAR)) {
            printUsage();
        } else {
            extractMetaProjectLocation(arg);
        }
    }

    private void extractSaveInterval(String s) {
        if (s.startsWith(SAVE_INTERVAL_OPTION)) {
            String min = s.substring(SAVE_INTERVAL_OPTION.length());
            int seconds = Integer.valueOf(min).intValue();
            System.out.println("Save interval sec=" + seconds);
            if (seconds > 0) {
                _saveIntervalMsec = seconds * 1000;
            }
        } else {
            printUsage();
        }
    }

    protected void extractMetaProjectLocation(String s) {
        metaprojectURI = URIUtilities.createURI(s);
    }

    protected void printUsage() {
        System.out.println(
            "usage: java -cp protege.jar edu.stanford.smi.protege.server.Server [options] <metaproject>");
        System.out.println("\tOptions:");
        System.out.println("\t\t" + SAVE_INTERVAL_OPTION + "<nseconds>");
        System.out.println("\t\t\tSave any dirty projects every n minutes (only needed for file based projects)");
        System.out.println("\t\t" + NOPRELOAD_OPTION);
        System.out.println("\t\t\tDon't preload projects.");
        System.exit(-1);
    }

    public Server(String[] args, RMISocketFactory sf) throws RemoteException, IOException {
        parseArgs(args);
        RMISocketFactory.setSocketFactory(sf);
        initialize();
    }

    public void reinitialize() throws RemoteException {
        shutdown();
        initialize();
    }

    private void shutdown() {
        _nameToOpenProjectMap.clear();
        _projectToServerProjectMap.clear();
        _sessions.clear();
        _sessionToProjectsMap.clear();
        stopUpdateThread();
    }

    private void initialize() throws RemoteException {
        Collection errors = new ArrayList();
        Project project = Project.loadProjectFromURI(metaprojectURI, errors);
        if (!errors.isEmpty()) {
            throw new RuntimeException(errors.iterator().next().toString());
        }
        _systemKb = project.getKnowledgeBase();
        _projectCls = _systemKb.getCls("Project");
        _userCls = _systemKb.getCls("User");
        _nameSlot = _systemKb.getSlot("name");
        _passwordSlot = _systemKb.getSlot("password");
        _locationSlot = _systemKb.getSlot("location");
        bindName();
        dumpProjects(preload);
        startUpdateThread();
    }

    private void dumpProjects(boolean preload) {
        if (preload) {
            System.out.print("Loading ");
        }
        System.out.println("Available Projects:");
        Iterator i = getAvailableProjectNames(null).iterator();
        while (i.hasNext()) {
            String name = (String) i.next();
            System.out.println("\t" + name);
            if (preload) {
                createProject(name);
            }
        }
    }

    protected void bindName() throws RemoteException {
        try {
            String boundName = getLocalBoundName();
            getRegistry().rebind(boundName, this);
            _baseURI = new URI("rmi:" + "//" + getMachineName() + "/" + boundName);
        } catch (Exception e) {
            e.printStackTrace();
            if (e instanceof RemoteException) {
                throw (RemoteException) e;
            } else {
                throw new RemoteException(e.getMessage());
            }
        }
    }

    private String getMachineName() {
        String name;
        try {
            name = InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            e.printStackTrace();
            name = "localhost";
        }
        return name;
    }

    public RemoteSession openSession(String username, String userMachine, String password) {
        RemoteSession session = null;
        if (isValid(username, password)) {
            session = new Session(username, userMachine);
            _sessions.add(session);
        }
        return session;
    }

    public void closeSession(RemoteSession session) {
        _sessions.remove(session);
    }

    public RemoteServerProject openProject(String projectName, RemoteSession session) {
        ServerProject serverProject = null;
        Project p = getOrCreateProject(projectName);
        if (p != null) {
            serverProject = getServerProject(p);
            if (serverProject == null) {
                serverProject = createServerProject(projectName, p);
                addServerProject(p, serverProject);
            }
            recordConnection(session, serverProject);
        }
        return serverProject;
    }

    private void recordConnection(RemoteSession session, ServerProject project) {
        // Log.enter(this, "recordConnection", session, project);
        Collection projects = (Collection) _sessionToProjectsMap.get(session);
        if (projects == null) {
            projects = new ArrayList();
            _sessionToProjectsMap.put(session, projects);
        }
        projects.add(project);
        project.register(session);
    }

    private void recordDisconnection(RemoteSession session, RemoteServerProject project) {
        // Log.enter(this, "recordDisconnection", session, project);
        Collection projects = (Collection) _sessionToProjectsMap.get(session);
        projects.remove(project);
    }

    private ServerProject getServerProject(String projectName) {
        Project p = getProject(projectName);
        return (p == null) ? null : getServerProject(p);
    }

    private ServerProject createServerProject(String name, Project p) {
        ServerProject impl = null;
        try {
            impl = new ServerProject(this, getURI(name), p);
        } catch (RemoteException e) {
            Log.exception(e, this, "createServerProject", name, p);
        }
        return impl;
    }

    private URI getURI(String projectName) {
        String name = FileUtilities.urlEncode(projectName);
        return _baseURI.resolve(name);
    }

    public void disconnectFromProject(RemoteServerProject serverProject, RemoteSession session) {
        recordDisconnection(session, serverProject);
    }

    private ServerProject getServerProject(Project p) {
        return (ServerProject) _projectToServerProjectMap.get(p);
    }

    private void addServerProject(Project p, RemoteServerProject sp) {
        _projectToServerProjectMap.put(p, sp);
    }

    private Project getProject(String name) {
        return (Project) _nameToOpenProjectMap.get(name);
    }

    private Project getOrCreateProject(String name) {
        Project project = getProject(name);
        if (project == null) {
            project = createProject(name);
        }
        return project;
    }

    private Project createProject(String name) {
        Project project = null;
        Iterator i = _systemKb.getInstances(_projectCls).iterator();
        while (i.hasNext()) {
            Instance instance = (Instance) i.next();
            String projectName = (String) instance.getOwnSlotValue(_nameSlot);
            if (projectName.equals(name)) {
                String projectLocation = (String) instance.getOwnSlotValue(_locationSlot);
                projectLocation = localizeLocation(projectLocation);
                URI uri = URIUtilities.createURI(projectLocation);
                project = Project.loadProjectFromURI(uri, new ArrayList(), true);
                localizeProject(project);
                _nameToOpenProjectMap.put(name, project);
                break;
            }
        }
        return project;
    }

    private static String localizeLocation(String location) {
        if (File.separatorChar != '\\') {
            location = location.replace('\\', File.separatorChar);
        }
        return location;
    }

    private void localizeProject(Project project) {
        localizeKB(project.getKnowledgeBase());
        localizeKB(project.getInternalProjectKnowledgeBase());
    }

    private void localizeKB(KnowledgeBase kb) {
        FrameStore fs = new LocalizeFrameStoreHandler(kb).newFrameStore();
        kb.insertFrameStore(fs);
    }

    public Collection getAvailableProjectNames(RemoteSession session) {
        List names = new ArrayList();
        Iterator i = _systemKb.getInstances(_projectCls).iterator();
        while (i.hasNext()) {
            Instance instance = (Instance) i.next();
            String fileName = (String) instance.getOwnSlotValue(_locationSlot);
            File file = new File(fileName);
            if (file.exists() && file.isFile()) {
                String name = (String) instance.getOwnSlotValue(_nameSlot);
                names.add(name);
            } else {
                Log.warning("Missing project at " + fileName, this, "getAvailableProjectNames");
            }
        }
        Collections.sort(names);
        return names;
    }

    public Collection getCurrentSessions(String projectName, RemoteSession session) {
        Collection currentSessions;
        RemoteServerProject project = getServerProject(projectName);
        if (project == null) {
            currentSessions = Collections.EMPTY_LIST;
        } else {
            currentSessions = getCurrentSessions(project);
        }
        return currentSessions;
    }

    public Collection getCurrentSessions(RemoteServerProject project) {
        Collection sessions = new ArrayList();
        Iterator i = _sessionToProjectsMap.entrySet().iterator();
        while (i.hasNext()) {
            Map.Entry entry = (Map.Entry) i.next();
            Collection projects = (Collection) entry.getValue();
            if (projects.contains(project)) {
                Session session = (Session) entry.getKey();
                sessions.add(session);
            }
        }
        return sessions;
    }

    private boolean isValid(String name, String password) {
        boolean isValid = false;
        Iterator i = _systemKb.getInstances(_userCls).iterator();
        while (i.hasNext()) {
            Instance user = (Instance) i.next();
            String username = (String) user.getOwnSlotValue(_nameSlot);
            if (username.equals(name)) {
                String userpassword = (String) user.getOwnSlotValue(_passwordSlot);
                if (userpassword.equals(password)) {
                    isValid = true;
                    break;
                }
            }
        }
        return isValid;
    }

    public String toString() {
        return "Server";
    }

    private void startUpdateThread() {
        if (_saveIntervalMsec != NO_SAVE) {
            _updateThread = new Thread("Save Projects") {
                public void run() {
                    try {
                        while (_updateThread == this) {
                            sleep(_saveIntervalMsec);
                            update();
                        }
                    } catch (InterruptedException e) {
                    }
                }
            };
            _updateThread.setDaemon(true);
            _updateThread.start();
        }
    }

    private void update() {
        // Log.enter(this, "update");
        ///CLOVER:FLUSH
        Collection errors = new ArrayList();
        Iterator i = _projectToServerProjectMap.entrySet().iterator();
        while (i.hasNext()) {
            Map.Entry entry = (Map.Entry) i.next();
            Project project = (Project) entry.getKey();
            ServerProject serverProject = (ServerProject) entry.getValue();
            // Log.trace("checking " + project, this, "update");
            if (serverProject.isDirty()) {
                errors.clear();
                Log.trace("saving " + project, this, "update");
                project.save(errors);
                serverProject.setClean();
                dumpErrors(project, errors);
            }
        }
    }

    private void dumpErrors(Project p, Collection errors) {
        if (!errors.isEmpty()) {
            Log.error("Unable to save project " + p, this, "dumpErrors");
            Iterator i = errors.iterator();
            while (i.hasNext()) {
                Object o = i.next();
                System.out.println("\t" + o.toString());
            }
        }
    }

    private void stopUpdateThread() {
        _updateThread = null;
    }
}
