/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.ui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.util.*;

/**
 * Panel to select a set of instances from all of the instances of a given set of classes.
 *
 * @author    Ray Fergerson <fergerson@smi.stanford.edu>
 */
public class SelectInstancesPanel extends JComponent {
    private static final int WIDTH = 500;
    protected JTree _clsTree;
    protected JList _instanceList;
    protected JComboBox _directAllInstanceComboBox;

    // static to remember the state across dialog openings.
    private static Object _oldDirectAllInstancesState = DirectInstancesList.DIRECT;

    protected SelectInstancesPanel(KnowledgeBase kb, Collection clses) {
        setPreferredSize(new Dimension(WIDTH, 300));
        if (clses.isEmpty()) {
            clses = kb.getRootClses();
        }
        createWidgets(kb, clses);
        fixRenderer();
    }

    protected LabeledComponent createClsesLabeledComponent(KnowledgeBase kb, Collection clses) {
        LabeledComponent clsesComponent = new LabeledComponent("Allowed Classes", new JScrollPane(_clsTree));
        clsesComponent.setFooterComponent(new ClsTreeFinder(kb, _clsTree, "Find Class"));
        return clsesComponent;
    }

    protected JComponent createClsTree(Collection clses) {
        LazyTreeRoot root = new ParentChildRoot(clses);
        _clsTree = ComponentFactory.createSelectableTree(null, root);
        _clsTree.addTreeSelectionListener(new TreeSelectionListener() {
            public void valueChanged(TreeSelectionEvent event) {
                loadInstances();
            }
        });
        FrameRenderer renderer = FrameRenderer.createInstance();
        renderer.setDisplayDirectInstanceCount(true);
        _clsTree.setCellRenderer(renderer);
        if (clses.size() == 1) {
            _clsTree.expandRow(0);
            _clsTree.setSelectionRow(0);
        }
        return _clsTree;
    }

    private void fixRenderer() {
        boolean displayType = _directAllInstanceComboBox.getSelectedItem().equals(DirectInstancesList.ALL);
        FrameRenderer frameRenderer = (FrameRenderer) _instanceList.getCellRenderer();
        frameRenderer.setDisplayType(displayType);
    }

    protected LabeledComponent createInstanceLabeledComponent() {
        LabeledComponent c = new LabeledComponent(null, new JScrollPane(_instanceList));
        c.setHeaderComponent(createDirectAllInstanceComboBox(), BorderLayout.WEST);
        c.setFooterComponent(new ListFinder(_instanceList, "Find Instance"));
        return c;
    }

    protected JComboBox createDirectAllInstanceComboBox() {
        _directAllInstanceComboBox = ComponentFactory.createComboBox();
        _directAllInstanceComboBox.addItem(DirectInstancesList.DIRECT);
        _directAllInstanceComboBox.addItem(DirectInstancesList.ALL);
        loadState();
        _directAllInstanceComboBox.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent event) {
                loadInstances();
                fixRenderer();
                saveState();
            }
        });
        return _directAllInstanceComboBox;
    }

    private void saveState() {
        _oldDirectAllInstancesState = _directAllInstanceComboBox.getSelectedItem();
    }
    private void loadState() {
        _directAllInstanceComboBox.setSelectedItem(_oldDirectAllInstancesState);
    }

    protected JComponent createInstanceList() {
        _instanceList = ComponentFactory.createList(null);
        _instanceList.setCellRenderer(FrameRenderer.createInstance());
        _instanceList.addMouseListener(new ModalDialogCloseDoubleClickAdapter());
        return _instanceList;
    }

    protected void createWidgets(KnowledgeBase kb, Collection clses) {
        createInstanceList();
        LabeledComponent instancesComponent = createInstanceLabeledComponent();

        createClsTree(clses);
        LabeledComponent clsesComponent = createClsesLabeledComponent(kb, clses);

        JSplitPane main = ComponentFactory.createLeftRightSplitPane();
        main.setLeftComponent(clsesComponent);
        main.setRightComponent(instancesComponent);
        main.setDividerLocation(WIDTH / 2);
        setLayout(new BorderLayout());
        add(main);
    }

    protected SimpleListModel getInstanceModel() {
        return (SimpleListModel) _instanceList.getModel();
    }

    public Collection getSelection() {
        return ComponentUtilities.getSelection(_instanceList);
    }

    protected void loadInstances() {
        ArrayList instances = new ArrayList();
        Iterator i = ComponentUtilities.getSelection(_clsTree).iterator();
        while (i.hasNext()) {
            Cls cls = (Cls) i.next();
            instances.addAll(getInstances(cls));
        }
        Collections.sort(instances, new FrameComparator());
        getInstanceModel().setValues(instances);
        if (!instances.isEmpty()) {
            _instanceList.setSelectedIndex(0);
        }
    }

    protected Collection getInstances(Cls cls) {
        boolean direct = _directAllInstanceComboBox.getSelectedItem().equals(DirectInstancesList.DIRECT);
        return (direct) ? cls.getDirectInstances() : cls.getInstances();
    }
}
