/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.util;

import java.awt.*;
import java.awt.dnd.*;
import java.util.*;
import java.util.List;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.tree.*;

/**
 * Default implementation of the various renderer interfaces.  This renderer is used for all Protege widget rendering.
 * This ensures a uniform look when a frame is displayed in a tree, a list, and a table.  This renderer also handles
 * the drag over feedback needed to supplement Java's pitiful drag and drop support.
 *
 * @author    Ray Fergerson <fergerson@smi.stanford.edu>
 */
public class DefaultRenderer extends JComponent implements TreeCellRenderer, ListCellRenderer, TableCellRenderer {
    public final static Object DROP_TARGET = new Object();
    public final static Object DROP_TARGET_AREA = new Object();
    public final static Object DROP_TARGET_AREA_ON = new Object();
    public final static Object DROP_TARGET_AREA_BELOW = new Object();
    public final static Object DROP_TARGET_AREA_ABOVE = new Object();

    protected Color _backgroundNormalColor;
    protected Color _foregroundNormalColor;
    protected Color _backgroundSelectionColor;
    protected Color _foregroundSelectionColor;
    private static final int ICON_TEXT_GAP = 3;
    private static final int TEXT_ICON_GAP = 0;
    private static final int ICON_ICON_GAP = 0;
    private static final int TEXT_TEXT_GAP = 0;
    protected FontMetrics _fontMetrics;
    protected LookAndFeel _cachedLookAndFeel;
    protected boolean _hasFocus;
    protected boolean _isSelected;
    protected boolean _grayedText;
    protected boolean _grayedSecondaryText = true;

    protected List _elements = new ArrayList();

    protected Object _dropTargetArea;
    protected Color _dropTargetLineColor = Color.black;

    // there should be a better way to cache these
    protected Color _treeForegroundSelectionColor;
    protected Color _treeBackgroundSelectionColor;
    protected Color _treeForegroundNormalColor;
    protected Color _treeBackgroundNormalColor;

    public DefaultRenderer() {
        loadTreeColors();
    }

    protected void checkDropTarget(Component component, Object value) {
        DropTarget d = component.getDropTarget();
        if (d != null && d.isActive()) {
            Object o = ((JComponent) component).getClientProperty(DROP_TARGET);
            if (equals(o, value)) {
                _dropTargetArea = ((JComponent) component).getClientProperty(DROP_TARGET_AREA);
            } else {
                _dropTargetArea = null;
            }
        }
    }

    protected void setFontStyle(int style) {
        setFont(getFont().deriveFont(style));
    }

    protected Color getBackgroundColor() {
        return (_isSelected) ? _backgroundSelectionColor : _backgroundNormalColor;
    }

    public Component getListCellRendererComponent(
        JList list,
        Object value,
        int row,
        boolean selected,
        boolean hasFocus) {
        _foregroundSelectionColor = list.getSelectionForeground();
        _backgroundSelectionColor = list.getSelectionBackground();
        _foregroundNormalColor = list.getForeground();
        _backgroundNormalColor = list.getBackground();
        return setup(list, value, hasFocus, selected);
    }

    public Dimension getPreferredSize() {
        Dimension d = new Dimension(0, _fontMetrics.getHeight());
        Object previousElement = null;
        Iterator i = _elements.iterator();
        while (i.hasNext()) {
            Object element = i.next();
            d.width += getGap(previousElement, element);
            updateSize(d, element);
            previousElement = element;
        }
        return d;
    }

    private int getGap(Object previousElement, Object currentElement) {
        int gap;
        if (previousElement == null) {
            gap = 0;
        } else if (previousElement instanceof Icon) {
            if (currentElement instanceof Icon) {
                gap = ICON_ICON_GAP;
            } else {
                gap = ICON_TEXT_GAP;
            }
        } else {
            if (currentElement instanceof Icon) {
                gap = TEXT_ICON_GAP;
            } else {
                gap = TEXT_TEXT_GAP;
            }
        }
        return gap;
    }

    private void updateSize(Dimension d, Object element) {
        if (element instanceof Icon) {
            updatePreferredSize(d, (Icon) element);
        } else if (element instanceof String) {
            updatePreferredSize(d, (String) element);
        } else {
            Log.error("Strange thing: " + element, this, "updateSize");
        }
    }

    private void updatePreferredSize(Dimension d, Icon icon) {
        d.width += icon.getIconHeight();
        d.height = Math.max(d.height, icon.getIconHeight());
    }

    private void updatePreferredSize(Dimension d, String text) {
        d.width += _fontMetrics.stringWidth(text);
    }

    public Component getTableCellRendererComponent(
        JTable table,
        Object value,
        boolean selected,
        boolean hasFocus,
        int row,
        int col) {
        _foregroundSelectionColor = table.getSelectionForeground();
        _backgroundSelectionColor = table.getSelectionBackground();
        _foregroundNormalColor = table.getForeground();
        _backgroundNormalColor = table.getBackground();
        return setup(table, value, hasFocus, selected);
    }

    protected Color getTextColor() {
        Color textColor;
        if (_isSelected) {
            textColor = _foregroundSelectionColor;
        } else {
            textColor = (_grayedText) ? Color.darkGray : _foregroundNormalColor;
        }
        return textColor;
    }

    public Component getTreeCellRendererComponent(
        JTree tree,
        Object value,
        boolean selected,
        boolean expanded,
        boolean leaf,
        int row,
        boolean hasFocus) {
        _foregroundSelectionColor = _treeForegroundSelectionColor;
        _backgroundSelectionColor = _treeBackgroundSelectionColor;
        _foregroundNormalColor = _treeForegroundNormalColor;
        _backgroundNormalColor = _treeBackgroundNormalColor;
        return setup(tree, value, hasFocus, selected);
    }

    public void load(Object o) {
        addText(o.toString());
    }

    public void loadNull() {
        addText("<null>");
    }

    protected void loadTreeColors() {
        _treeForegroundSelectionColor = UIManager.getColor("Tree.selectionForeground");
        _treeBackgroundSelectionColor = UIManager.getColor("Tree.selectionBackground");
        _treeForegroundNormalColor = UIManager.getColor("Tree.textForeground");
        _treeBackgroundNormalColor = UIManager.getColor("Tree.textBackground");
    }

    public void paint(Graphics graphics) {
        graphics.setFont(getFont());
        Dimension preferredSize = getPreferredSize();
        Dimension currentSize = getSize();

        graphics.setColor(getBackgroundColor());
        graphics.fillRect(0, 0, currentSize.width, currentSize.height);
        graphics.setColor(getTextColor());

        Point position = new Point(0, (currentSize.height - preferredSize.height) / 2);
        Object previousElement = null;
        boolean paintedString = false;
        Iterator i = _elements.iterator();
        while (i.hasNext()) {
            Object currentElement = i.next();
            position.x += getGap(previousElement, currentElement);
            if (currentElement instanceof Icon) {
                paintIcon(graphics, (Icon) currentElement, position, currentSize);
            } else if (currentElement instanceof String) {
                Color color = null;
                if (paintedString && _grayedSecondaryText) {
                    color = Color.gray;
                }
                paintString(graphics, (String) currentElement, position, color);
                paintedString = true;
            }
            previousElement = currentElement;
        }

        if (position.x > currentSize.width) {
            String dots = "...";
            int dotsWidth = _fontMetrics.stringWidth(dots);
            graphics.setColor(getBackgroundColor());
            graphics.fillRect(currentSize.width - dotsWidth, position.y, dotsWidth, currentSize.height);
            graphics.setColor(getTextColor());
            graphics.drawString(dots, currentSize.width - dotsWidth, position.y + _fontMetrics.getAscent());
        }

        if (_dropTargetArea == DROP_TARGET_AREA_ON) {
            graphics.setColor(_dropTargetLineColor);
            graphics.drawRect(0, 0, position.x - 1, currentSize.height - 1);
        } else if (_dropTargetArea == DROP_TARGET_AREA_BELOW) {
            graphics.setColor(_dropTargetLineColor);
            int y = currentSize.height - 1;
            graphics.drawLine(0, y, currentSize.width - 1, y);
        } else if (_dropTargetArea == DROP_TARGET_AREA_ABOVE) {
            graphics.setColor(_dropTargetLineColor);
            graphics.drawLine(0, 0, currentSize.width - 1, 0);
        } else {
            // do nothing
        }
    }

    protected void paintIcon(Graphics graphics, Icon icon, Point position, Dimension size) {
        int y = position.y + (size.height - icon.getIconHeight()) / 2;
        icon.paintIcon(this, graphics, position.x, y);
        position.x += icon.getIconWidth();
    }

    protected void paintString(Graphics graphics, String text, Point position) {
        paintString(graphics, text, position, null);
    }

    protected void paintString(Graphics graphics, String text, Point position, Color color) {
        if (color != null) {
            graphics.setColor(color);
        }
        graphics.drawString(text, position.x, position.y + _fontMetrics.getAscent());
        position.x += _fontMetrics.stringWidth(text);
    }

    protected void setForegroundColorOverride(Color color) {
        _foregroundNormalColor = color;
        _foregroundSelectionColor = color;
    }

    public void setGrayedText(boolean b) {
        _grayedText = b;
    }

    public void setGrayedSecondaryText(boolean b) {
        _grayedSecondaryText = b;
    }

    public void setInvalid() {
    }

    public void setMainIcon(Icon icon) {
        if (_elements.isEmpty()) {
            if (icon != null) {
                _elements.add(icon);
            }
        } else if (_elements.get(0) instanceof Icon) {
            if (icon == null) {
                _elements.remove(0);
            } else {
                _elements.set(0, icon);
            }
        } else {
            if (icon != null) {
                _elements.add(0, icon);
            }
        }
    }

    public String getMainText() {
        return (String) getFirst(String.class);
    }

    public Icon getMainIcon() {
        return (Icon) getFirst(Icon.class);
    }

    public int getIconTextGap() {
        return ICON_TEXT_GAP;
    }

    public int getTextIconGap() {
        return TEXT_ICON_GAP;
    }

    public int getTextTextGap() {
        return TEXT_TEXT_GAP;
    }

    public int getIconIconGap() {
        return ICON_ICON_GAP;
    }

    public Object getFirst(Class clas) {
        Object first = null;
        Iterator i = _elements.iterator();
        while (i.hasNext()) {
            Object o = i.next();
            if (clas.isInstance(o)) {
                first = o;
                break;
            }
        }
        return first;
    }

    public void setMainText(String text) {
        if (_elements.isEmpty()) {
            _elements.add(text);
        } else if (_elements.get(0) instanceof String) {
            _elements.set(0, text);
        } else if (_elements.size() == 1) {
            _elements.add(1, text);
        } else {
            _elements.set(1, text);
        }
    }

    public void appendIcon(Icon icon) {
        addIcon(icon);
    }

    public void appendText(String text) {
        addText(text);
    }

    public void addIcon(Icon icon) {
        _elements.add(icon);
    }

    public void addText(String text) {
        _elements.add(text);
    }

    protected void setNormalForegroundColor(Color color) {
        _foregroundNormalColor = color;
    }

    protected Component setup(Component c, Object value, boolean hasFocus, boolean isSelected) {
        _grayedText = false;
        Font font = c.getFont();
        if (font.isBold()) {
            font = font.deriveFont(Font.PLAIN);
        }
        setFont(font);
        _hasFocus = hasFocus;
        _isSelected = isSelected;

        _elements.clear();
        if (value == null) {
            loadNull();
        } else if (value instanceof LazyTreeNode) {
            load(((LazyTreeNode) value).getUserObject());
        } else {
            load(value);
        }

        _fontMetrics = getFontMetrics(getFont());
        LookAndFeel currentLookAndFeel = UIManager.getLookAndFeel();
        if (currentLookAndFeel != _cachedLookAndFeel) {
            loadTreeColors();
            _cachedLookAndFeel = currentLookAndFeel;
        }
        checkDropTarget(c, value);
        return this;
    }

    public void clear() {
        _elements.clear();
    }

    public static boolean equals(Object o1, Object o2) {
        return SystemUtilities.equals(o1, o2);
    }

    // overridden for performance reasons to do nothing
    public boolean isOpaque() {
        return true;
    }
    public void validate() {
    }
    public void revalidate() {
    }
    public void invalidate() {
    }
    public void repaint() {
    }
    public void repaint(int x, int y, int w, int h) {
    }
    public void repaint(Rectangle r) {
    }
    public void firePropertyChange(String propertyName, boolean oldValue, boolean newValue) {
    }
    public void firePropertyChange(String propertyName, byte oldValue, byte newValue) {
    }
    public void firePropertyChange(String propertyName, char oldValue, char newValue) {
    }
    public void firePropertyChange(String propertyName, double oldValue, double newValue) {
    }
    public void firePropertyChange(String propertyName, float oldValue, float newValue) {
    }
    public void firePropertyChange(String propertyName, int oldValue, int newValue) {
    }
    public void firePropertyChange(String propertyName, long oldValue, long newValue) {
    }
    public void firePropertyChange(String propertyName, short oldValue, short newValue) {
    }
    public void firePropertyChange(String propertyName, Object oldValue, Object newValue) {
    }
}
