/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.util;

import java.awt.*;
import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.nio.charset.*;
import java.util.*;
import java.util.List;

import javax.swing.*;
import javax.swing.plaf.*;
import javax.swing.plaf.metal.*;

import edu.stanford.smi.protege.plugin.*;
import edu.stanford.smi.protege.resource.*;
import edu.stanford.smi.protege.server.*;

/**
 * A set of utilities for accessing the underlying system and for manipulating system level objects.
 *
 * @author  Ray Fergerson <fergerson@smi.stanford.edu>
 * @author  Joe Edelman (jxe@dartmouth.edu)
 */
public class SystemUtilities {

    static {
        init();
    }

    public static void initialize() {
        // do nothing except ensure that the static initializer has run
    }
    public static void debugBreak() {
    }

    public static void exit() {
        try {
            System.exit(0);
        } catch (SecurityException e) {
            // do nothing
        }
    }

    public static void beep() {
        Toolkit.getDefaultToolkit().beep();
    }

    public static Class forName(String className) {
        return PluginUtilities.forName(className);
    }

    public static Class forName(String className, boolean promiscuous) {
        return PluginUtilities.forName(className, promiscuous);
    }

    public static void gc() {
        // Log.enter(SystemUtilities.class, "gc");
        System.gc();
        System.runFinalization();
        System.gc();
    }

    /*
    private static String getLaunchString(String document) {
        String launchString;
        if (isWindowsNtOrBetter()) {
            launchString = "cmd /c start " + document;
        } else if (isWindows()) {
            launchString = "start " + document;
        } else {
            String urlString;
            if (document.startsWith("http")) {
                urlString = document;
            } else {
                urlString = FileUtilities.toURL(new File(document)).toString();
            }
            String browser = ApplicationProperties.getBrowser();
            launchString = browser + " " + urlString;
        }
        return launchString;
    }
    */

    public static String getSystemProperty(String property) {
        String value;
        try {
            value = System.getProperty(property);
        } catch (SecurityException e) {
            value = null;
        }
        return value;
    }

    public static String getSystemProperty(String property, String defaultValue) {
        String value;
        try {
            value = System.getProperty(property, defaultValue);
        } catch (SecurityException e) {
            value = defaultValue;
        }
        return value;
    }

    public static String getUserDirectory() {
        return getSystemProperty("user.dir");
    }

    public static String getMachineName() {
        String machineName;
        try {
            machineName = InetAddress.getLocalHost().getCanonicalHostName();
        } catch (UnknownHostException e) {
            machineName = "Unknown";
        }
        return machineName;
    }

    private static String encoding;
    public static String getFileEncoding() {
        if (encoding == null) {
            String historicalName = new OutputStreamWriter(new ByteArrayOutputStream()).getEncoding();
            encoding = Charset.forName(historicalName).name();
        }
        return encoding;
    }

    public static String getUserName() {
        return getSystemProperty("user.name");
    }

    private static void init() {
        try {
            printSystemInfo(System.out);
            setSecurityManager();
            PluginUtilities.initialize();
            loadLookAndFeel();
        } catch (Exception e) {
            // We explicitly do nothing fancy with writing this output.  This method is called on startup
            // and there may be problems with almost anything (i.e. System.out). This is the best chance we
            // have to getting a reasonable error message.
            e.printStackTrace();
        }
    }

    private static void setSecurityManager() {
        try {
            System.setSecurityManager(new LaxSecurityManager());
        } catch (SecurityException e) {
            // expect this in applets
        }
    }

    private static void loadLookAndFeel() {
        MetalLookAndFeel.setCurrentTheme(new ProtegeMetalTheme());
        String className = ApplicationProperties.getLookAndFeelClassName();
        if (className == null) {
            // need to force a ui update to install the theme
            try {
                UIManager.setLookAndFeel(UIManager.getLookAndFeel());
            } catch (Exception e) {
                Log.exception(e, SystemUtilities.class, "loadLookAndFeel");
            }
        } else {
            try {
                UIManager.setLookAndFeel(className);
            } catch (Exception e) {
                Log.exception(e, SystemUtilities.class, "loadLookAndFeel");
            }
        }
    }

    private static class ProtegeMetalTheme extends DefaultMetalTheme {
        private FontUIResource fontResource;
        //        private ColorUIResource secondaryColorResource1;
        //        private ColorUIResource secondaryColorResource2;
        //        private ColorUIResource secondaryColorResource3;
        //        private static final int LEVEL = 244;

        public FontUIResource getControlTextFont() {
            if (fontResource == null) {
                fontResource = new FontUIResource(super.getControlTextFont().deriveFont(Font.PLAIN));
            }
            return fontResource;
        }
        public FontUIResource getMenuTextFont() {
            if (fontResource == null) {
                fontResource = new FontUIResource(super.getMenuTextFont().deriveFont(Font.PLAIN));
            }
            return fontResource;
        }
        //        public ColorUIResource getSecondary1() {
        //            if (secondaryColorResource1 == null) {
        //                secondaryColorResource1 = new ColorUIResource(new Color(LEVEL - 100, LEVEL - 100, LEVEL - 100));
        //            }
        //            return secondaryColorResource1;
        //        }
        //        public ColorUIResource getSecondary2() {
        //            if (secondaryColorResource2 == null) {
        //                secondaryColorResource2 = new ColorUIResource(new Color(LEVEL - 50, LEVEL - 50, LEVEL - 50));
        //            }
        //            return secondaryColorResource2;
        //        }
        //        public ColorUIResource getSecondary3() {
        //            if (secondaryColorResource3 == null) {
        //                secondaryColorResource3 = new ColorUIResource(new Color(LEVEL, LEVEL, LEVEL));
        //            }
        //            return secondaryColorResource3;
        //        }
    }

    public static boolean isWindows() {
        return getSystemProperty("os.name").indexOf("Windows") != -1;
    }

    public static boolean isMac() {
        return getSystemProperty("os.name").indexOf("Mac") != -1;
    }

    public static boolean modalDialogInDropWorks() {
        // Bringing up a modal dialog during a drop only appears to work on Windows with JVM 1.3.
        // It is broken on JDK1.2 and in JDK 1.4 and
        // on the Mac with JDK 1.3.  This I am going to just disable it until I can figure out
        // how to make it work realiably.
        return false;
    }

    public static Object newInstance(Class clas, Class[] argumentClasses, Object[] arguments) {
        Object instance = null;
        try {
            Constructor constructor = clas.getConstructor(argumentClasses);
            instance = constructor.newInstance(arguments);
        } catch (Throwable e) {
            Log.exception(e, SystemUtilities.class, "newInstance", clas, argumentClasses, arguments);
        }
        return instance;
    }

    public static Object newInstance(String className) {
        Object o = null;
        try {
            Class clas = forName(className);
            if (clas == null) {
                Log.warning("no such class", SystemUtilities.class, "newInstance", className);
            } else {
                o = clas.newInstance();
            }
        } catch (Throwable e) {
            Log.exception(e, SystemUtilities.class, "newInstance", className);
        }
        return o;
    }

    public static Object newInstance(Class clas) {
        Object o = null;
        try {
            o = clas.newInstance();
        } catch (Throwable e) {
            Log.exception(e, SystemUtilities.class, "newInstance", clas);
        }
        return o;
    }

    public static void pause() {
        try {
            System.out.flush();
            System.err.flush();
            System.out.print("Press <Enter> to continue");
            System.in.read();
            while (System.in.available() != 0) {
                System.in.read();
            }
        } catch (Exception e) {
        }
    }

    public static void printMemoryUsage() {
        gc();
        printMemoryUsageNoGC();
    }

    public static void printMemoryUsageNoGC() {
        Runtime runtime = Runtime.getRuntime();
        long total = runtime.totalMemory();
        long free = runtime.freeMemory();
        String s = "memory: total=" + total + ", used=" + (total - free);
        Log.trace(s, SystemUtilities.class, "printMemoryUsage");
    }

    public static void printSystemInfo(PrintStream stream) {
        StringBuffer buffer = new StringBuffer();
        buffer.append(Text.getProgramTextName());
        buffer.append(" ");
        buffer.append(Text.getVersion());
        buffer.append(" ");
        buffer.append(Text.getBuildInfo());
        buffer.append(", JVM ");
        buffer.append(getSystemProperty("java.runtime.version"));
        buffer.append(", ");
        buffer.append(getSystemProperty("os.name"));
        buffer.append(", encoding=");
        buffer.append(getFileEncoding());
        buffer.append(", language=");
        buffer.append(Locale.getDefault().getLanguage());
        buffer.append(", country=");
        buffer.append(Locale.getDefault().getCountry());
        stream.println(buffer.toString());
        stream.flush();
        // printSystemProperties(stream);
    }

    private static class PropertyComparator implements Comparator {
        public int compare(Object o1, Object o2) {
            Map.Entry entry1 = (Map.Entry) o1;
            Map.Entry entry2 = (Map.Entry) o2;
            return entry1.getKey().toString().compareToIgnoreCase(entry2.getKey().toString());
        }
    }

    public static void printSystemProperties(PrintStream stream) {
        stream.println("System Properties:");
        List entries = new ArrayList(System.getProperties().entrySet());
        Collections.sort(entries, new PropertyComparator());
        Iterator i = entries.iterator();
        while (i.hasNext()) {
            Map.Entry entry = (Map.Entry) i.next();
            String key = entry.getKey().toString();
            if (!key.startsWith("lax.")) {
                stream.println("\t" + key + "=" + entry.getValue());
            }
        }
    }

    public static void showHTML(String url) {
        try {
            if (!url.startsWith("http:") && !url.startsWith("file:")) {
                url = new File(url).toURI().toURL().toString();
            }
            // Log.enter(SystemUtilities.class, "showHTML", url);
            BrowserLauncher.openURL(url);
        } catch (IOException e) {
            Log.exception(e, SystemUtilities.class, "showHTML", url);
        }
    }

    public static void sleepMsec(int msecs) {
        try {
            Thread.sleep(msecs);
        } catch (Exception e) {
            Log.exception(e, SystemUtilities.class, "sleepMsec", new Integer(msecs));
        }
    }

    public static Boolean toBoolean(Object o) {
        Boolean b;
        if (o instanceof Boolean) {
            b = (Boolean) o;
        } else {
            String s = o.toString();
            if (s.equalsIgnoreCase("true")) {
                b = Boolean.TRUE;
            } else if (s.equalsIgnoreCase("false")) {
                b = Boolean.FALSE;
            } else {
                b = null;
            }
        }
        return b;
    }

    public static Float toFloat(Object o) {
        Float f;
        if (o instanceof Float) {
            f = (Float) o;
        } else {
            try {
                f = Float.valueOf(o.toString());
            } catch (Exception e) {
                f = null;
            }
        }
        return f;
    }

    public static Integer toInteger(Object o) {
        Integer i;
        if (o instanceof Integer) {
            i = (Integer) o;
        } else {
            try {
                i = Integer.valueOf(o.toString());
            } catch (Exception e) {
                i = null;
            }
        }
        return i;
    }

    public static boolean equals(Object o1, Object o2) {
        return (o1 == null) ? o2 == null : o1.equals(o2);
    }

    public static Collection getClassesWithAttribute(String key, String value) {
        return PluginUtilities.getClassesWithAttribute(key, value);
    }

}
